/*****************************************************************************
     This file is part of the Changedfiles system. 

     Copright (C)2002 Philip Antoniades

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*******************************************************************************/

#include "action_factory.h"
#include "action.h"
#include "command_action.h"
// #include "action_exception.h"
#include "ftp_action.h"
#include "exec_action.h"
#include "mysql_action.h"
#include "operations.h"
#include <pthread.h> 
#include <time.h>
#include <stdio.h>
#include <string> 


using std::string; 


ActionFactory* ActionFactory::m_action_factory = NULL; 

namespace 
{
  pthread_mutex_t m_action_factory_mutex = PTHREAD_MUTEX_INITIALIZER; 
  pthread_mutex_t m_action_factory_other_mutex = PTHREAD_MUTEX_INITIALIZER; 
  pthread_mutex_t m_action_factory_open_mutex = PTHREAD_MUTEX_INITIALIZER; 
} 

ActionFactory::ActionFactory()
{ 
  m_action_factory_instantiated = true; 
}

ActionFactory::~ActionFactory()
{ 

}

ActionFactory * ActionFactory::get_instance()
{ 
  pthread_mutex_lock( &m_action_factory_mutex ); 
  if (  m_action_factory == NULL || m_action_factory->is_instantiated() == false ) 
    { 
      m_action_factory = new ActionFactory(); 
    } 
  pthread_mutex_unlock( &m_action_factory_mutex ); 
  return m_action_factory; 
}

/*
 * parse the Rule's action string and create the proper action.
 */
Action * ActionFactory::create_action( FileOperation fo, Rule* rule )
{
  string strrule = rule->get_rule_str( fo.get_operation() );
  e_action_type type = parse_type( strrule );
  e_optype optype = fo.get_operation();
  Action* action;
  ExecAction* e_action;
  
  this->substitute_keywords( strrule, &fo );
  
  try
    {
      switch( type ){
      case ActionFactory::FTP:
	action = new FTPAction( optype, strrule, rule->get_retry() );
	break;
#ifdef _CHANGEDFILES_HAS_MYSQL
      case ActionFactory::MYSQL:
	action = new MySQLAction( strrule, rule->get_retry() );
	break;
#endif			    
      case ActionFactory::EXEC:
	e_action = new ExecAction( strrule, rule->get_retry());			
	e_action->set_operation( fo.get_operation() );
	if (  fo.get_operation() ==  SYMLINK
	      || fo.get_operation() == LINK
	      || fo.get_operation() == RENAME )
	  {
	    e_action->set_old_file( fo.get_old_file() );
	    e_action->set_new_file( fo.get_new_file() );
	  }
	else
	  {    	
	    e_action->set_file( fo.get_filename() );
	  }
	action = e_action;
	break;
      case ActionFactory::COMMAND:
	// just drop through to default
      default:
	action = new CommandAction( strrule, rule->get_retry() );
      }
      action->set_timeout( rule->get_timeout() );
      action->set_command_timeout( rule->get_command_timeout() );
      char message[500];
      Logger* logger = Logger::get_instance();
      sprintf( message, "action_factory.cpp: setting action timeout_retry to %d", rule->get_timeout_retry() );
      logger->log( Logger::DEBUG, message);
      action->set_timeout_retry( rule->get_timeout_retry() );
      action->set_operation( fo );
    }
  catch( ... )
    {
      // why bother, but ok
      action = NULL;
      throw;
    }

  return action;
}

/** 
 * Pulls out the first word to determine Action Type
 * and for the actual types it removes the keyword.
 */
ActionFactory::e_action_type ActionFactory::parse_type( string& line )
{
  string::size_type pos; 
  ActionFactory::e_action_type type;

  // trim left side of spaces...
  pos = line.find_first_not_of( " \t" );
  line = line.substr( pos );

  pos = line.find_first_of( " \t" ); 
  if ( pos == string::npos ) 
    { 
      //      throw ActionException( "No valid tokens found in action line" );
      // TODO report the error
    }
  string token = line.substr( 0, pos ); 

  // we requre all tokens in uppercase
  if( token == "FTP" || token == "PFTP" )
    {
      type = ActionFactory::FTP;
      line = line.substr( pos );
    }
  else if ( token == "EXEC" )
    {
      type = ActionFactory::EXEC;
      line = line.substr( pos );
    }
  else if ( token == "SHELL" )
    {
      type = ActionFactory::COMMAND;
      line = line.substr( pos );
    }
#ifdef _CHANGEDFILES_HAS_MYSQL
  else if ( token == "MYSQL" )
    {
      type = ActionFactory::MYSQL;
      line = line.substr( pos );
    }
#endif
  else
    { 
      type = ActionFactory::COMMAND;
    }

  return type;
}

bool ActionFactory::is_instantiated()
{ 
  return m_action_factory_instantiated; 
}

/**
 * changes __FILE__ to <filename>
 * a miserable hack of mike's code.
 */
void ActionFactory::substitute_keywords( string& line, FileOperation* fo )
{
  Logger* logger = Logger::get_instance();
  string::size_type pos; 
  switch ( fo->get_operation() )
    {
    case SYMLINK:
    case LINK:
    case RENAME: 			
      while ( ( pos = line.find( old_file_keyword )) != string::npos )
	{ 
	  line.replace( pos, old_file_keyword.length(), fo->get_old_file() );
	}
      while ( ( pos = line.find( new_file_keyword )) != string::npos )
	{ 
	  line.replace( pos, new_file_keyword.length(), fo->get_new_file() );
	}
      break;
    	
    case OPENR:
    case MKDIR:
    case CREATE:
    case RMDIR:
    case UNLINK:
    case CLOSE:
    default:
      while ( ( pos = line.find( filename_keyword )) != string::npos )
	{ 
	  line.replace( pos, filename_keyword.length(), fo->get_filename() );
	}
    }
  
  logger->log( Logger::DEBUG, 
	       "debug - ActionFactory::substitute_keywords() returning " 
	       + line); 

  return;
}
  
