/*****************************************************************************
     This file is part of the Changedfiles system. 

     Copright (C)2001  Michael L. Welles

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*******************************************************************************/
/*  $Id: command_action.cpp,v 1.12 2003/09/20 22:15:07 philip Exp $ */
#include "command_action.h"
#include "logger.h"
#include <string>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h> 
extern "C" {
#include <time.h>
#include <signal.h>
}

using std::string; 

namespace {
  const char* SHELL = "/bin/sh"; 

};

CommandAction::CommandAction( string action, int max_retry )
{
  m_command = action;
  m_max_retry = max_retry;
  this->set_command_timeout( 0 );
  this->set_timeout_retry( 0 );
}

CommandAction::CommandAction( string action, int max_retry, time_t command_timeout )
{ 
  m_command = action; 
  m_max_retry = max_retry; 
  this->set_command_timeout( command_timeout );
}


CommandAction::CommandAction( string action, int max_retry, time_t command_timeout, int timeout_retry )
{ 
  m_command = action; 
  m_max_retry = max_retry; 
  this->set_command_timeout( command_timeout );
  this->set_timeout_retry( timeout_retry );
}


bool CommandAction::set_command( string action )
{ 
  m_command = action; 
  return true; 
} 

string CommandAction::get_command() 
{ 
  return m_command; 
} 

string CommandAction::get_description(){
  return "Command: " + m_command;
}


bool CommandAction::do_action() //throw( ActionException )
{
  int status;
  pid_t pid; 
  string msg;
  Logger* logger = Logger::get_instance();
  bool ret = false;
  if ( ( pid = fork() ) ) 
    {
      bool done = false; 
      time_t start_time = time( NULL );
      time_t now_time;
      while ( ! done ) 
	{ 
	  if ( waitpid( pid, &status, WUNTRACED | WNOHANG ) != 0 )
	    { 
	      done = true;
	    }
	  else
	    { 
	      if ( this->get_command_timeout() > 0 ) 
		{ 
		  now_time = time( NULL );
		  if ( now_time > (start_time + this->get_command_timeout() ))
		  {
		    char message[500];

		    sprintf( message, "Command timed out, %d seconds elapsed.  Killing PID %d.", 
			     (int) this->get_command_timeout(), (int) pid);
		    logger->log( Logger::ERR, message );
		    kill( pid, SIGKILL );
		    waitpid( pid, &status, 0 );
		    if ( ! this->add_retry() )
		      { 
			ret = true; 
			msg = "Retry count exceeded for " + this->get_description();
			logger->log( Logger::ERR, msg.c_str() );
		      }
		    else 
		      { 
			ret = false; 
		      }
		    return ret;
		  }
		}
	      usleep( 200 );
	    }
	}
    }		     
  else 
    { 
      string action = this->get_command();
  
      if( execl( SHELL, SHELL, "-c", action.c_str(), NULL ) ){
	ret = true;
      }else{
	//should never get here
	exit(1); 
      }

    }

  if ( WIFEXITED( status ) ) 
    { 
      if ( WEXITSTATUS( status ) != 0 )
	{ 
	  msg = "error - Action failed, non-zero exit status. ";  
	  msg += this->get_description();
	  logger->log( Logger::ERR, msg ); 
	}
      else 
	{ 
	  msg = "info  - Action succeeded:  " + this->get_description(); 
	  logger->log( Logger::INFO, msg.c_str() ); 
	  ret = true;
	}
    }
  else if ( WIFSIGNALED( status ) )
    { 
      msg = "error - Action failed, unhandled signal passed to child.  "; 
      msg += this->get_description();
      logger->log( Logger::ERR, msg.c_str() ); 
      ret = false;
    }
  else if ( WIFSTOPPED( status ) )
    { 
      msg = "error - Action stopped: "; 
      msg += this->get_description();
      logger->log( Logger::ERR, msg.c_str() ); 
      ret = false;
    }
  else 
    { 
      msg = "error - Action failed, abnormal exit: "; 
      msg += this->get_description();
      logger->log( Logger::ERR, msg.c_str() ); 
      ret = false;
    }

  // we return success if we've exceeded our retry limit.
  // so we get removed from the queue and deleted.
  if( ! ret )
    {
      if ( ! this->add_retry() )
	{ 
	  ret = true;
	  msg = "Retry count exceeded for " + this->get_description();
	  logger->log( Logger::ERR, msg.c_str() );
	}
    }

  return ret;
}
