/*****************************************************************************
     This file is part of Changedfiles

     Copright (C)2002  Mike Welles

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*******************************************************************************/

#include "ftp_action.h"
#include "logger.h"
// #include "action_exception.h"

extern "C"{
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <libfget.h>
#include <signal.h>
}

#include <exception>
#include <stdexcept>
namespace 
{
  void alarm_handler( int sig )
  {
    Logger* logger = Logger::get_instance();
    logger->log( Logger::ERR, "ALARM received, ftp command action timed out");
  }
}

FTPAction::FTPAction( e_optype optype, string params, int max_retry) //throw( ActionException )
{
  m_optype = optype;
  this->parse_params( params );
  this->set_max_retry( max_retry );
  // these could all be setable...
  // well, this one is. but DEFAULT
  this->set_timeout( -1 );
  this->set_timeout_retry( 0 );
  this->m_cache_expire = -1;
  this->m_flags = FTP_PASSIVE;
}

FTPAction::FTPAction( e_optype optype, string params, int max_retry, int timeout, int timeout_retry) //throw ( ActionException )
{
  m_optype = optype;
  this->parse_params( params );
  this->set_max_retry( max_retry );
  // these could all be setable...
  // well, this one is. but DEFAULT
  this->set_timeout( timeout);
  this->set_timeout_retry( timeout_retry );
  this->m_cache_expire = -1;
  this->m_flags = FTP_PASSIVE;
}


/** 
 * Note: `ftp -h` says:
 * 	Usage: { ftp | pftp } [-pinegvtd] [hostname]
 *	   -p: enable passive mode (default for ftp and pftp)
 *	   -i: turn off prompting during mget
 *	   -n: inhibit auto-login
 *	   -e: disable readline support, if present
 *         -g: disable filename globbing
 *	   -v: verbose mode
 *	   -t: enable packet tracing [nonfunctional]
 *	   -d: enable debugging
 * hardly helpful to these purposes, so the param string will be 
 * ftp -u username -p password -h <host> -r <remote dir> -l <local dir> _FILE_ 
 * for now.
 */
bool FTPAction::parse_params( string params ) // throw (ActionException)
{
  //int numparams = 0;
  string::size_type pos = 0;
  string::size_type space;
  string paramcopy;
  Logger* logger = Logger::get_instance();

  string p = "FTPAction: ";
  // get all the cmd line switches
  while( ( pos = params.find(" -", 0 ) ) != string::npos )
    {
      paramcopy = params.substr( (pos + 2) );
      
      char op = paramcopy.at( 0 );
      
      // find first character not of any in this string...
      pos = paramcopy.find_first_not_of( " \t", 1 );
      
      space = paramcopy.find_first_of( " \t", pos );
      
      if( space == string::npos )
	{
	  logger->log( Logger::ERR, "File name not supplied for ftp action" );
	  //throw ActionException( "File name not supplied" );
	  return false;
	}
      
      // now we need to do a 'trim left'
      // and then substr( pos, find_next_space )
      string opvalue = paramcopy.substr( pos, (space - pos ) );
      
      switch( op ){
      case 'u':
      case 'U':
	this->m_username = opvalue;
	p = p + " user:" + this->m_username;
	break;
      case 'p':
      case 'P':
	this->m_password = opvalue;
	p = p + " password: xxxxxxxx";
	break;
      case 'r':
	this->m_remote_dir = opvalue;
	p = p + " remote:" + opvalue;
	  break;
      case 'l':
	this->m_local_dir = opvalue;
	p = p + " local:" + opvalue;
	break;
      case 'h':
      case 'H':
	this->m_hostname = opvalue;
	p = p + " hostname:" + opvalue;
	break;
      default:
	Logger* logger = Logger::get_instance();
	string unknown = "Unknown paramater in ftp command string - ";
	unknown += op;
	logger->log( Logger::ERR, unknown );
      }

      params = paramcopy.substr( space );
      
    } // end while
  
  
  
  // trim left side of spaces...
  pos = params.find_first_not_of( " \t", 1 );
  params = params.substr( pos );
  
  // now parse the filename and host
  // look for two file names (if we're a rename, we can use it)
  // 
  space = params.find_last_of( " \t" );
  string::size_type inner_space = params.find_first_of( " \t" );
  if ( (inner_space != string::npos) && (space == inner_space) )
    {
      this->m_oldname = params.substr( 0, inner_space++ );
      params = params.substr( inner_space );
      // trim left side of spaces...
      pos = params.find_first_not_of( " \t" );
      params = params.substr( pos );
      
      // we do renames by attempting to delete the remote "orignal" file.
      // no point in doing this unless we have some expectation
      // that it might be found on the remote machine.
      
      pos = this->m_oldname.find( this->m_local_dir );
      if ( pos != 0 ) 
	{
	  this->m_oldname = "";
	}
      else
	{
	  int sz = this->m_local_dir.size();
	  sz++;
	  string oldbasename = this->m_oldname.substr( sz );
	  this->m_oldname = this->m_remote_dir; 
	  this->m_oldname += "/";
	  this->m_oldname += oldbasename;
	}
    }
  else
    {
      this->m_oldname = "";
      
    }
  
  
  space = params.find_last_of( " \t" );
  
  if( space == string::npos ){ 
    this->m_filename = params.substr( 0 );
  }else{
    this->m_filename = params.substr( 0, space++ );
  }
  
  pos = this->m_filename.find( this->m_local_dir );
  if ( pos == 0 ) 
    {
      int sz = this->m_local_dir.size();
      sz++;
      this->m_basename = this->m_filename.substr( sz );
      // 	if ( this->m_oldname.size() > 0 ) 
      // 	  { 
      // 	    this->m_oldname = this->m_oldname.substr( sz );
      // 	  }
    }
  else
    {
      this->m_basename = this->m_filename;
    }
  
  //    this->m_hostname = params.substr( space );
  //  p += " local filename: ";
  //     p += this->m_filename;
  //     p += " basename: ";
  //     p += this->m_basename;
  //     p += " remote filename: ";
  //     p += this->m_remote_dir;
  //     p += "/";
  //     p += this->m_basename;
  //     p += " remote oldname: ";
  //     p += this->m_oldname;
  //     logger->log( Logger::INFO, p );
  return true;
}

/**
 * make the necessary replacements & omissions for 
 * path names on the remote host
 */ 
string FTPAction::get_remote_filename()// throw (ActionException)
{
  if( this->m_remote_dir.length() < 1 && this->m_local_dir.length() < 1 ){
    return this->m_filename;
  }
  string remote = this->m_remote_dir;
  remote += this->m_basename;
  return remote;
}

bool FTPAction::connect()
{

  FTPConnectionManager *ftp_manager = FTPConnectionManager::get_instance();
  
  m_ftp = ftp_manager->get_connection( this->m_hostname, 
				       this->m_username, 
				       this->m_password );
  if ( m_ftp == NULL ) 
      return false; 
  return true;
}


/** 
 * this cant throw exceptions. 
 * it's gotta plow through.
 */
bool FTPAction::disconnect()
{
  bool ret = false;
  if( m_ftp == NULL ) {
    return true;
  }
  FTPConnectionManager *ftp_manager = FTPConnectionManager::get_instance();

  ret = ftp_manager->return_connection( m_hostname, m_username, m_password, m_ftp );
  m_ftp = NULL;
  return ret;
}

bool FTPAction::send_file( string remote_name )
{
  FTPFILE* ftpfile;
  int filefd;
  size_t buf_size = 1024;
  char buffer[buf_size];
  string errStr;

  try
    {
      filefd = open( this->m_filename.c_str(), O_RDONLY );

      if ( filefd < 0 )
        { 
	  errStr = "Error opening file ";
	  errStr += this->m_filename;
	  errStr += " during ftp action";
	  errStr += ": ";
	  errStr +=  strerror( errno );
	  //throw ActionException( errStr );
	  Logger *logger = Logger::get_instance();
	  logger->log( Logger::ERR, errStr );
	  return false; 
        }

      if( ftp_open( &ftpfile, m_ftp, (char *)remote_name.c_str(), O_WRONLY ) )
	{
	  errStr = strerror( errno );
	  //throw ActionException( "FTP File Open unsuccessful: " + errStr );
	  Logger *logger = Logger::get_instance();
	  string message = "file open unsuccessful in ftp action ";
	  message += errStr;
	  logger->log( Logger::ERR, message );
	  if ( ftpfile != NULL )
	    { 
	      ftp_close( ftpfile );
	    }
	  close( filefd );
	  return false; 
	}

      int have_read = 0;

      while ( ( have_read = read(  filefd,  buffer, buf_size ) ) > 0 )
        { 
	  int written = ftp_write( ftpfile, buffer,  have_read );
	  if ( written < 0 )
	    { 
	      errStr = strerror( errno );
	      // throw ActionException( "FTP File write error: " + errStr );
	      // we should never get here
	      Logger* logger = Logger::get_instance();
	      string message = "file write error in ftp action: ";
	      message += errStr;
	      logger->log( Logger::ERR, message );
	      if ( ftpfile != NULL ) 
		{
		  ftp_close( ftpfile );
		}
	      close( filefd );
	      return false;
	      break;
	    }
        }
    }
  catch( ... )
    {
      Logger* logger = Logger::get_instance();
      try
	{
	  close( filefd );
	}
      catch( ... )
	{ 
	  logger->log( Logger::ERR, "Close of files failed in ftp action!" );
	}
      logger->log( Logger::ERR, "Error sending file!" );
      throw;
    }

  close( filefd );

  if( ftp_close( ftpfile ) )
    {
      errStr = strerror( errno );
      Logger *logger = Logger::get_instance();
      string message = "file close unsucessful in ftp action: ";
      message += errStr;
      logger->log( Logger::ERR, message );
      //throw ActionException( "FTP File Close unsuccessful: " + errStr );
      return false;
    }
  return true;
}

bool FTPAction::mkdir( string remote_name )
{
  if ( ftp_mkdir( m_ftp, (char *)remote_name.c_str() ) == -1 )
    {
      return false;
    }
  return true;
}

bool FTPAction::rmdir( string remote_name )
{
  if ( ftp_rmdir( m_ftp, (char *)remote_name.c_str() ) == -1 )
    {
      return false;
    }
  return true;
}

bool FTPAction::unlink( string remote_name )
{
  if ( ftp_remove( m_ftp, (char *)remote_name.c_str() ) == -1 )
    { 
      return false;
    }
  return true;;
}

bool FTPAction::rename_file( string remote_name )
{
  if ( send_file( remote_name ) )
    {
      if ( this->m_oldname.size() > 0 ) 
	{
	  unlink( this->m_oldname );
	}
      return true;
    }
  return false;
}





/** 
 * should trap & handle all ftp-related (network) fexceptions
 * and ONLY throw ActionExceptions
 */
bool FTPAction::do_action() //throw( ActionException )
{
  bool ret = true;
  Logger* logger = Logger::get_instance();
  string str_remote = this->get_remote_filename();
  /// FIXME -- if exception handling ever works right with gcc, then 
  // remove the test and retry stuff, and allow the later exception 
  // catching bit to handle it.
  if ( ! this->connect() )
    { 
      if( this->add_retry() > 0 )
	{
	  return false;
	}
      else
	{
	  logger->log( Logger::ERR, "Retries exceeded for ftp action." );
	  return true; 
	}
    }
  
  if ( get_command_timeout() > 0 ) 
    {
      signal( SIGALRM, alarm_handler );
      alarm( get_command_timeout() );
    }
  switch( m_optype )
    { 
    case OPENW:
      if ( this->send_file( str_remote ) )
	{
	  logger->log( Logger::INFO, "SENT " + m_filename + " to " 
		       + m_hostname + ":" + str_remote);
	}
      else 
	{
	  logger->log( Logger::ERR, "SEND FAILED " + m_filename + " to " 
		       + m_hostname + ":" + str_remote);
	  ret = false;
	}
      break;
    case OPENR:
      break;
    case MKDIR:
      if ( this->mkdir( str_remote ))
	{
	  logger->log( Logger::INFO, "CREATED DIRECTORY " + m_hostname + ":" + str_remote );
	}
      else
	{
	  logger->log( Logger::ERR, "DIRECTORY CREATION FAILED " +  m_hostname + ":" + str_remote );
	  ret = false;
	}
      break;
    case RMDIR:
      if ( this->rmdir( str_remote ))
	{
	  logger->log( Logger::INFO, "REMOVED DIRECTORY " + m_hostname + ":" + str_remote );
	}
      else
	{
	  logger->log( Logger::ERR, "DIRECTORY REMOVAL FAILED " + m_hostname + ":" + str_remote );
	  ret = false;
	}
      break;
    case UNLINK:
      if ( this->unlink( str_remote ))
	{ 
	  ;
	  logger->log( Logger::INFO, "REMOVED "  + m_hostname + ":" + str_remote );
	}
      else
	{
	  logger->log( Logger::ERR, "REMOVAL FAILED "  + m_hostname + ":" + str_remote  );
	  ret = false;
	}
      break;
    case SYMLINK:
      // Judgement call: if we have a symlink rule for FTP
      // we assume that the rule creator understands that
      // you can't symlink over ftp, and wants to send
      // the link as a file -- even if it means it'll
      // be stale should the original change.
      if ( this->send_file( str_remote ) )
	{
	  logger->log( Logger::INFO, "SENT " + m_filename + " to "  + m_hostname + ":" + str_remote  );
	}
      else 
	    {
	      logger->log( Logger::ERR, "FAILED TO SEND " + m_filename + " to "  + m_hostname + ":" + str_remote  );
	      ret = false;
	    }
      break;
    case CREATE:
      if ( this->send_file( str_remote ) )
	{
	  logger->log( Logger::INFO, "SENT " + m_filename + " to "  + m_hostname + ":" + str_remote  );
	}
      else 
	{
	  logger->log( Logger::ERR, "FAILED TO SEND " + m_filename + " to "  + m_hostname + ":" + str_remote  );
	  ret = false;
	}
      break;
    case LINK:
      // see judgment call for SYMLINK above
      if ( this->send_file( str_remote ) )
	{
	  logger->log( Logger::INFO, "SENT " + m_filename + " to "  + m_hostname + ":" + str_remote  );
	}
      else 
	{
	  logger->log( Logger::ERR, "FAILED TO SEND " + m_filename + " to " + m_hostname + ":" + str_remote  );
	  return false;
	  ret = false;
	}
      break;
    case CLOSE:
      // noop
      break;
    case RENAME:
      if ( this->rename_file( str_remote ) )
	{ 
	  logger->log( Logger::INFO, "RENAMED " + m_filename + " to "  + m_hostname + ":" + str_remote  );
	}
      else
	{
	  logger->log( Logger::ERR, "FAILED TO RENAME " + m_filename + " to "  + m_hostname + ":" + str_remote  );
	  ret = false;
	}
      break;
    case CHMOD:
      logger->log( Logger::ERR, "Error: CHMOD is not a supported action for ftp rules.\n");
      break;

    case CHOWN:
      logger->log( Logger::ERR, "Error: CHOWN is not a supported action for ftp rules.\n");
      break;

    case NOOP:
      // noop
      break;
    default:
      //noop
      break;
      ;
    }
  alarm( 0 );
  this->disconnect();
  if ( ret == false )
    { 
      if( this->add_retry() > 0 )
	{
	  ret = false;
	}
      else
	{
	  logger->log( Logger::ERR, "Retries exceeded for ftp action. Deleting file." );
	  ret = true; 
	}
    }
  return ret;
}

string FTPAction::get_description()
{
  string desc = "Send file " + this->m_filename + " to FTP host " 
    + this->m_hostname + ":" + this->get_remote_filename();

  return desc;
}
