/*****************************************************************************
     This file is part of Changedfiles

     Copright (C)2003  Michael Welles

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*******************************************************************************/
#ifdef _CHANGEDFILES_HAS_MYSQL_
#include "mysql_action.h"
#include "mysql_connection_manager.h"
#include "logger.h"
// #include "action_exception.h"

extern "C"{
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <libfget.h>
#include <signal.h>
}

#include <exception>
#include <stdexcept>
namespace 
{
  void alarm_handler( int sig )
  {
    Logger* logger = Logger::get_instance();
    logger->log( Logger::ERR, "ALARM received, mysql command action timed out");
  }
}

MySQLAction::MySQLAction( string params, int max_retry) //throw( ActionException )
{
  this->parse_params( params );
  this->set_max_retry( max_retry );
  // these could all be setable...
  // well, this one is. but DEFAULT
  this->set_timeout( -1 );
  this->set_timeout_retry( 0 );
}

MySQLAction::MySQLAction( string params, int max_retry, int timeout, int timeout_retry) //throw ( ActionException )
{
  this->parse_params( params );
  this->set_max_retry( max_retry );
  // these could all be setable...
  // well, this one is. but DEFAULT
  this->set_timeout( timeout);
  this->set_timeout_retry( timeout_retry );
}


bool MySQLAction::parse_params( string params ) // throw (ActionException)
{
  //int numparams = 0;
  string::size_type pos = 0;
  string::size_type space;
  string paramcopy;
  Logger* logger = Logger::get_instance();

  string p = "MySQLAction: ";
  // get all the cmd line switches
  while( ( pos = params.find(" -", 0 ) ) != string::npos )
    {
      paramcopy = params.substr( (pos + 2) );
      
      char op = paramcopy.at( 0 );
      
      // find first character not of any in this string...
      pos = paramcopy.find_first_not_of( " \t", 1 );
      
      space = paramcopy.find_first_of( " \t", pos );
      
      if( space == string::npos )
	{
	  logger->log( Logger::ERR, "File name not supplied for ftp action" );
	  //throw ActionException( "File name not supplied" );
	  return false;
	}
      
      // now we need to do a 'trim left'
      // and then substr( pos, find_next_space )
      string opvalue = paramcopy.substr( pos, (space - pos ) );
      
      switch( op ){
      case 'u':
      case 'U':
	this->m_username = opvalue;
	p = p + " user:" + this->m_username;
	break;
      case 'p':
      case 'P':
	this->m_password = opvalue;
	p = p + " password: xxxxxxxx";
	break;
      case 'h':
      case 'H':
	this->m_hostname = opvalue;
	p = p + " hostname:" + opvalue;
	break;
      case 'd':
      case 'D':
	this->m_database = opvalue;
	p = p + " database: " + opvalue;
	break;
      default:
	Logger* logger = Logger::get_instance();
	string unknown = "Unknown paramater in MySQL command - ";
	unknown += op;
	logger->log( Logger::ERR, unknown );
      }

      params = paramcopy.substr( space );
      
    } // end while
  
  
  
  // trim left side of spaces...
  pos = params.find_first_not_of( " \t", 1 );
  params = params.substr( pos );
  
  this->m_query = params;
  return true;
}


bool MySQLAction::do_action() //throw( ActionException )
{
  bool ret = true;
  Logger* logger = Logger::get_instance();
   /// FIXME -- if exception handling ever works right with gcc, then 
  // remove the test and retry stuff, and allow the later exception 
  // catching bit to handle it.
   signal( SIGALRM, alarm_handler );
   if ( get_command_timeout() > 0 )
     {
       alarm( get_command_timeout() );
     }
   ret = do_query();
   if ( ret == false )
     { 
       if( this->add_retry() > 0 )
	 {
	   ret = false;
	 }
       else
	 {
	   logger->log( Logger::ERR, "Retries exceeded for mysql action.  Abandoning." );
	   ret = true; 
	 }
     }
   return ret;
}

bool MySQLAction::do_query()
{
  Logger *logger=Logger::get_instance();
  MySQLConnectionManager 
    *cmanager = MySQLConnectionManager::get_instance();
    MYSQL *mysql = cmanager->get_connection( m_hostname.c_str(), 
					     m_username.c_str(),
					     m_password.c_str(),
					     m_database.c_str());
    MYSQL_RES *result;
    unsigned int num_fields;
    unsigned int num_rows;
    string info = "";
    if ( mysql == NULL )
      {
	return false;
      }

    if ( mysql_query( mysql, m_query.c_str() ))
      {
	string error = "Error execucting sql in MySQL query: ";
	error += mysql_error( mysql );
	logger->log( Logger::ERR, error );
	return false;
      }
    
    result = mysql_store_result(mysql);
    
    if (result)
      {
	num_fields = mysql_num_fields(result);
	info = "MySQL query successful. ";
	info += num_fields;
	info += " fields affected.";
	mysql_free_result(result);
      }
    else 
      {
	if ( mysql_field_count( mysql ) == 0 )
	  {
	    num_rows = mysql_affected_rows(mysql);
	    info = "MySQL query successful. ";
	    info += num_rows;
	    info += " rows affected.";
	  }
	else
	  {
	    info = "MySQL query failed: ";
	    info += mysql_error( mysql );
	    logger->log( Logger::ERR, info );
	    return false;
	  }
      }
    logger->log( Logger::INFO, info );
    cmanager->return_connection( m_hostname.c_str(),
				 m_username.c_str(),
				 m_password.c_str(),
				 m_database.c_str(),
				 mysql );
    return true;
}
#endif
