/*****************************************************************************
     This file is part of the Changedfiles system. 

     Copright (C)2001  Michael L. Welles

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*******************************************************************************/
/*  $Id: rule.cpp,v 1.13 2003/12/01 23:51:09 mike Exp $ */
#include "rule.h"
#include "operations.h" 
#include "file_operation.h"
#include <string> 
#include <regex.h>
#include <list> 

using std::string; 

namespace 
{ 
  const string filename_keyword = "__FILE__"; 
  const string new_file_keyword = "__NEW_FILE__"; 
  const string old_file_keyword = "__OLD_FILE__"; 
} 


Rule::Rule()
{ 
  m_logger = Logger::get_instance(); 
  m_command_timeout = 0;
  m_timeout_retry = 0;
}

bool Rule::set_rule_str( const string rule_str ) 
{ 
  set_name( rule_str );
  m_rule_str = rule_str; 
  return true; 
}

bool Rule::set_regexp( string regexp_str )
{ 
  regex_t preg; 
  regcomp( &preg, regexp_str.c_str(), REG_EXTENDED ); 
  return set_regexp( preg ); 
}

bool Rule::set_regexp( regex_t regexp ) 
{ 
  m_regexp = regexp; 
  return true; 
}

regex_t Rule::get_regexp()
{ 
  return m_regexp; 
}

bool Rule::set_retry( int retry )
{ 
  m_retry = retry; 
  return true; 
} 

int Rule::get_retry()
{ 
  return m_retry; 
}

bool Rule::set_timeout( int timeout )
{
  m_timeout = timeout; 
  return true;
}

int Rule::get_timeout()
{ 
  return m_timeout; 
} 

bool Rule::set_command_timeout( int timeout )
{
  m_command_timeout = timeout;
  return true;
}

int Rule::get_command_timeout()
{
  return m_command_timeout;
}

bool Rule::set_rule_str( e_optype op, const string command )
{ 
  switch ( op ) 
    { 
    case OPENW: 
      m_openw_rule = command; 
      break; 

    case OPENR:
      m_openr_rule = command; 
      break; 

    case MKDIR: 
      m_mkdir_rule = command; 
      break; 

    case RMDIR: 
      m_rmdir_rule = command; 
      break; 
      
    case UNLINK:
      m_unlink_rule = command; 
      break; 
      
    case SYMLINK: 
      m_symlink_rule = command; 
      break; 
      
    case LINK: 
      m_link_rule = command; 
      break; 

    case CLOSE: 
      m_close_rule = command; 
      break; 

    case RENAME: 
      m_rename_rule = command; 
      break;

    case CHOWN:
      m_chown_rule = command;
      break;

    case CHMOD:
      m_chmod_rule = command;
      break;

    default: 
      return false; 
    }
  return true; 
}

string Rule::get_rule_str( e_optype op )
{
  switch( op )
    {
      
    case SYMLINK: 
      return m_symlink_rule;
      break; 
      
    case LINK: 
      return m_link_rule;
      break; 

    case RENAME: 
      return m_rename_rule;
      break; 
      
    case OPENW: 
      return m_openw_rule;
      break; 

    case OPENR:
      return m_openr_rule;
      break; 

    case MKDIR: 
      return m_mkdir_rule;
      break; 

    case RMDIR: 
      return m_rmdir_rule;
      break; 
      
    case UNLINK:
      return m_unlink_rule;
      break; 

    case CLOSE: 
      return m_close_rule;
      break;

    case CHMOD:
      return m_chmod_rule;
      break;

    case CHOWN:
      return m_chown_rule;
      break;

    default: 
      return ""; 
    }
  return ""; 
}
	
bool Rule::operation_matches( FileOperation* file ){
  e_optype op =  file->get_operation();

  if( file->get_filename().length() )
    {
      return operation_matches( op, file->get_filename() );
    }
  else
    {
      if ( operation_matches( op, file->get_old_file() ))
	{
	  if ( ! is_to_be_ignored( file->get_new_file() ))
	    {
	      return true; 
	    }
	  else
	    {
	      return false;
	    }
	}
      if ( operation_matches( op, file->get_new_file() ))
	{
	  if ( ! is_to_be_ignored( file->get_old_file() ))
	    {
	      return true;
	    }
	  else
	    {
	      return false;
	    }
	}
    }
  return false; 
}

bool Rule::operation_matches ( e_optype op, const string filename )
{ 
  m_logger = Logger::get_instance();

  if( filename.length() < 1 ){
    m_logger->log( Logger::ERR, "No filename passed to Rule::operation_matches" );
  }

  if ( this->file_matches( filename ) )
    { 
      string msg = "debug - " + filename + " matches rule " + m_rule_str + " and has a command for "; 
      switch ( op ) 
	{ 
	case OPENW: 
	  if ( m_openw_rule.length() ) 
	    { 
	      msg += "OPENW"; 
	      m_logger->log( Logger::DEBUG, msg ); 
	      return true;
	    }
	  break; 
	  
	case OPENR:
	  if ( m_openr_rule.length() )
	    { 
	      msg += "OPENR"; 
	      m_logger->log( Logger::DEBUG, msg ); 
	      return true; 
	    }
	  break; 
	  
	case MKDIR: 
	  if ( m_mkdir_rule.length() ) 
	    { 
	      msg += "MKDIR"; 
	      m_logger->log( Logger::DEBUG, msg ); 
	      return true; 
	    }
	  break; 
	  
	case RMDIR: 
	  if ( m_rmdir_rule.length() ) 
	    { 
	      msg += "RMDIR"; 
	      m_logger->log( Logger::DEBUG, msg ); 
	      return true; 
	    }
	  break; 
	  
	case UNLINK:
	  if ( m_unlink_rule.length() ) 
	    { 
	      msg += "UNLINK"; 
	      m_logger->log( Logger::DEBUG, msg ); 
	      return true; 
	    }
	  break; 
	  
	case SYMLINK: 
	  if ( m_symlink_rule.length() ) 
	    { 
	      msg += "SYMLINK"; 
	      m_logger->log( Logger::DEBUG, msg ); 
	      return true; 
	    }
	  break; 
	  
	case LINK: 
	  if ( m_link_rule.length() ) 
	    { 
	      msg += "LINK"; 
	      m_logger->log( Logger::DEBUG, msg ); 
	      return true; 
	    }
	  break; 
	  
	case CLOSE: 
	  if ( m_close_rule.length() ) 
	    { 
	      msg += "CLOSE"; 
	      m_logger->log( Logger::DEBUG, msg ); 
	      return true; 
	    }
	  break; 
	  
	case RENAME: 
	  if ( m_rename_rule.length() ) 
	    { 
	      msg += "RENAME"; 
	      m_logger->log( Logger::DEBUG, msg ); 
	      return true; 
	    }
	  break; 

	case CHMOD:
	  if ( m_chmod_rule.length() )
	    {
	      msg += "CHMOD";
		m_logger->log( Logger::DEBUG, msg );
	      return true;
	    }
	  break;

	case CHOWN:
	  if (m_chown_rule.length() )
	    {
	      msg += "CHOWN";
	      m_logger->log( Logger::DEBUG, msg );
	      return true;
	    }
	  break;

	default: 
	  return false; 
	}
    }
  return false; 
} 
			
bool Rule::file_matches ( const string filename )
{
  if ( regexec( &m_regexp, filename.c_str(), 0, NULL, 0 ) == 0 )
    { 
      for ( IGNORE_LIST::iterator i = ignore_list.begin(); 
          i != ignore_list.end(); ++i )
      { 
        regex_t reg = *i; 
      if ( regexec( &reg, filename.c_str(), 0, NULL, 0 ) == 0 )
	{ 
	  return false; 
	}
      }
      return true; 
    } 
  return false; 
}

bool Rule::is_to_be_ignored( const string filename )
{
  for ( IGNORE_LIST::iterator i = ignore_list.begin(); 
	i != ignore_list.end(); ++i )
    { 
      regex_t reg = *i; 
      if ( regexec( &reg, filename.c_str(), 0, NULL, 0 ) == 0 )
	{ 
	  return true; 
	}
    }
  return false; 
}

bool Rule::add_ignore( string regexp_str )
{ 
  regex_t preg; 
  regcomp( &preg, regexp_str.c_str(), REG_EXTENDED ); 
  ignore_list.push_back( preg ); 
  return true;
}



      
      
      
